package vishnu.server;

import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import vishnu.util.Misc;
import vishnu.util.Sammon;
import vishnu.cluster.*;
import vishnu.datablock.*;

public class CKServer 
{
    private String index_directory = null;

    private final static String FILE_DOC_FREQ = "doc_freq";
    private final static String FILE_TERM_FREQ = "term_freq";
    private final static String FILE_CK_WORDS = "ck_words";
    
    private float[] docFreq = null;
    private String[] ck_words = null;
    private Vector termFreq = null;


    private int[] x_keywords;
    private int[] x_hitdocs;
    private float[][] KDMatrix;
    private Vector[] clusters;
    private double[][] centroids;
    private Point2D[] sammon;
    private Vector descriptions;
    
    public CKServer(String collection_home, String view) {
	
	this.index_directory = collection_home;
	if (!this.index_directory.endsWith(File.separator)) {
	    this.index_directory += File.separator;
	}
	this.index_directory += "index" + File.separator + "ck_index" + File.separator;
	if (view != null) {
	    this.index_directory += view + File.separator;
	}
	    
	loadIndices();
    }
    
    private void loadIndices(){
	FileInputStream fis = null;
	ObjectInputStream in = null;
	
	try {
	    docFreq = (float[])getObject(this.index_directory  + FILE_DOC_FREQ);
	    ck_words = (String[])getObject(this.index_directory + FILE_CK_WORDS);
	    termFreq = (Vector)getObject(this.index_directory + FILE_TERM_FREQ);
	} catch (Exception e) {
	    System.err.println("CKServer Error: Couldn't load up ck index data: "+e);
	}
    }

    
    private Object getObject(String str)
	throws IOException, FileNotFoundException, ClassNotFoundException {
	Object o = null;
	FileInputStream fis = null;
	ObjectInputStream in = null;
	fis = new FileInputStream(str);
	in = new ObjectInputStream(fis);
	o = in.readObject();
	in.close();
	return o;
    }


    public void setDescriptions(Vector desc)
    {
	descriptions = desc;
    }
    

    public String[] computeKeywords(Vector docNum)
    {
	int hits = docNum.size();
	
	Hashtable keyHash = new Hashtable();
	
	int count_ck_all = 0;
	
	for( int i = 0; i < hits; i++ ){
	    
	    int doc_index = ((Integer)docNum.elementAt(i)).intValue();
	    Hashtable ht = (Hashtable)termFreq.elementAt(doc_index);
	    count_ck_all += ht.size();
	    
	    Enumeration enum = ht.keys();
	    while( enum.hasMoreElements() ){
		Integer term_index = (Integer)enum.nextElement();
		
		Integer set_freq;
		if( (set_freq = (Integer)keyHash.get(term_index)) == null ){
		    keyHash.put(term_index,new Integer(1));
		}else{
		    keyHash.put(term_index,new Integer(set_freq.intValue()+1));
		}
	    }
	}
	

	/**** compute interesting weight for each keyword ****/
	
	Enumeration enum = keyHash.keys();
	
	float[] weight_ck = new float[keyHash.size()];
	Integer[] weight_index = new Integer[keyHash.size()];
	int count = 0;
	
	while( enum.hasMoreElements() ){
	    Integer index_ck = (Integer)enum.nextElement();
	    Integer set_freq = (Integer)keyHash.get(index_ck);
	    
	    float set_freq_f = (float)set_freq.intValue()/hits;
	    
	    float f1 = (float)set_freq_f/docFreq[index_ck.intValue()];
	    float f2 = -(float)Math.log(set_freq_f);
	    
	    weight_index[count] = index_ck;
	    weight_ck[count] = - f1 * f2;
	    
	    count++;
	}
	
	Misc.quickSort(weight_ck,weight_index);
	
	
	/**** store indices of up to 100 top ranked keywords ****/
	
	int wordCount = Math.min(100,count);
	int[] indices_kw = new int[wordCount];
	
	for( int i = 0; i < wordCount; i++ ){
	    indices_kw[i] = weight_index[i].intValue();
	    
	    float weight = -weight_ck[i];
	    String ck = ck_words[indices_kw[i]];
	}
	
	
	/**** sort the indices_kw array in increasing order of indices ****/
	
	Misc.quickSort(indices_kw);

	x_keywords = indices_kw;
	
	String[] keywords = new String[x_keywords.length];
	for( int i = 0; i < x_keywords.length; i++ ){
	    keywords[i] = ck_words[x_keywords[i]];
	}

	return keywords;
    }
    
    
    public SparseMatrix getSparseMatrix(Vector docNums){
		
	int hits = docNums.size();
	int wordCount = x_keywords.length;
	int scale = 100; // length of document vector
	

	/**** build full matrix: rows indices_kw (wordCount), col docNum (hits) ****/
	
	float[][] matrix_kd = new float[hits][wordCount];
	float[] sum_rows = new float[hits];
	int emptyRows = 0;
	
	for( int i = 0; i < hits; i++ ){
	    
	    int doc_index = ((Integer)docNums.elementAt(i)).intValue();
	    Hashtable ht = (Hashtable)termFreq.elementAt(doc_index);
	    
	    for( int j = 0; j < wordCount; j++ ){
		
		Integer term_freq;
		
		if( (term_freq = (Integer)ht.get(new Integer(x_keywords[j]))) != null ){
		    float w = -(float)(Math.log(1+term_freq.intValue()) * Math.log(docFreq[x_keywords[j]]));
		    w = term_freq.intValue();
		    matrix_kd[i][j] = w;
		    sum_rows[i] += w*w;
		}
	    }
	    
	    
	    /**** sqrt sum to obtain |z| ****/
	    
	    if( sum_rows[i] != 0 ) sum_rows[i] = (float)Math.sqrt(sum_rows[i])/scale; else emptyRows++;
	}
	
	/**** populate sparseMatrix and full matrix dividing each entry by the row_sum ****/
	
	SparseMatrix matrix = new SparseMatrix(hits-emptyRows,wordCount);
	KDMatrix = new float[hits-emptyRows][wordCount];
		
	int ci = 0;

	for( int i = hits-1; i >= 0; i-- ){
	    if( sum_rows[i] == 0 ){
		docNums.removeElementAt(i);
		descriptions.removeElementAt(i);
	    }
	}
	
	for( int i = 0; i < hits; i++ ){
	    
	    if( sum_rows[i] == 0 ) continue;
	    
	    for( int j = 0; j < wordCount; j++ ){
		
		float m = matrix_kd[i][j];
		
		if( m != 0 ){
		    float norm_m = m; //(float)m/sum_rows[i];
		    matrix.addDataItem(ci, j, norm_m);
		    KDMatrix[ci][j] = norm_m; //(float)(norm_m);
		}
		else KDMatrix[ci][j] = 0;
	    }
	    ci++;
	}
	
	x_hitdocs = new int[docNums.size()];
	
	for( int i = 0; i < docNums.size(); i++ ){
	    x_hitdocs[i] = ((Integer)docNums.elementAt(i)).intValue();
	}
	
	return matrix;
    }


    public Vector[] getClusters(){

	int wordCount = KDMatrix[0].length;
	int docHits = KDMatrix.length;
	System.err.println("getClusters: word count = "+wordCount+", docHits = "+docHits);
        int clusterNum =  docHits>20?10:docHits;
      
        int limit =  docHits/2;
	
        Clustering clustering = new Clustering(Clustering.COMPLETE_LINKAGE,KDMatrix,1,clusterNum,limit);
	
        clusters = clustering.getClusters();
	centroids= clustering.getCentroids();
	
	// what are these for??
	//Sammon map = new Sammon(centroids);
	//Point2D [] sammon = map.getMapping();
	return clusters;
    }
    
    
    public int[] getHitDocuments()
    {
	return x_hitdocs;
    }


    public double[][] getCentroids()
    {
	return centroids;
    }


    public String[] getHitDescriptions()
    {
	String[] desc = new String[descriptions.size()];
	for( int i = 0; i < desc.length; i++ ){
	    desc[i] = (String)descriptions.elementAt(i);
	}

	return desc;
    }
    
    public Point2D[] getSammonMap(double[][] centroids)
    {
	if (centroids == null) {
	    return null;
	}
	Sammon sammon = new Sammon(centroids);
	Point2D[] p = sammon.getMapping();
	
	return p;
    }
}
